import os
import pathlib
from gsc import cli
from gsc.exercises import (
    sync_error,
    merge_conflict,
    multiple_remotes,
    use_the_force,
    revert,
    reset_file,
    amend,
)


class ResetError(Exception):
    pass


def reset():
    while not os.path.exists(".git"):
        os.chdir("..")
        if os.getcwd() == "/":
            raise ResetError("This is not a git repo.")

    if not os.path.exists(".gsc_id"):
        raise ResetError("This is not a Git Scientist exercise.")
    gsc_id = pathlib.Path(".gsc_id").read_text().strip()

    cli.info(f"Resetting {gsc_id}")

    if gsc_id == "sync_error":
        sync_error.reset()
    if gsc_id == "merge_conflict":
        merge_conflict.reset()
    if gsc_id == "multiple_remotes":
        multiple_remotes.reset()
    if gsc_id == "use_the_force":
        use_the_force.reset()
    if gsc_id == "revert":
        revert.reset()
    if gsc_id == "reset_file":
        reset_file.reset()
    if gsc_id == "amend":
        amend.reset()
    else:
        raise ResetError("Unknown Git Scientist exercise. Try upgrading gsc.")



import json
import shutil
import pathlib
import subprocess

from gsc import verifier, cli, setup_exercise
from gsc.exercises import utils

UPSTREAM_URL = "https://github.com/git-scientist/fork-this"
UPSTREAM_URL_GIT = UPSTREAM_URL + ".git"


def setup():
    cli.info("There's no setup for this exercise.")


def reset():
    cli.info("There's no reset for this exercise.")


def verify():
    # We should have an origin remote with a fork url
    res = subprocess.run(["git", "remote", "-v"], capture_output=True)
    if res.returncode != 0:
        raise verifier.VerifyError("Failed to list remote repos")

    remotes = {}
    res = res.stdout.decode("utf-8").strip().split("\n")
    for r in res:
        [name, url] = r.split("\t")
        [url, _] = url.split(" ")
        remotes[name] = url

    if remotes["origin"] == UPSTREAM_URL:
        raise verifier.VerifyError(
            "You've cloned the upstream repo. You need to fork the repo first and then clone your fork."
        )

    num_remotes = len(remotes)
    if num_remotes == 1:
        raise verifier.VerifyError(
            "This repo only has one remote. You need to add the upstream repo as a second remote."
        )
    elif num_remotes > 2:
        raise verifier.VerifyError(
            f"This repo has {num_remotes} remotes. There should be exactly 2.\n"
            "Delete your local repo, clone your fork, and try adding the upstream remote again."
        )

    # We should have an upstream remote with the original url
    if "upstream" not in remotes:
        remote_names = list(remotes.keys())
        remote_names.remove("origin")
        [name] = remote_names
        if remotes[name] in [UPSTREAM_URL, UPSTREAM_URL_GIT]:
            cli.warn(
                f'I couldn\'t find the upstream remote, but I did find a remote called "{name}" with the correct url.\n'
                'You should consider calling your upstream remotes "upstream" instead.'
            )
        else:
            raise verifier.VerifyError(
                "I couldn't find the upstream remote.\n"
                f'I found a remote called "{name}" with the wrong url.\n'
                f"{UPSTREAM_URL} is the correct url.\n"
                f"{remotes[name]} is what you have."
            )
    elif remotes["upstream"] not in [UPSTREAM_URL, UPSTREAM_URL_GIT]:
        raise verifier.VerifyError(
            "The upstream url is not correct.\n"
            f"{UPSTREAM_URL} is the correct url.\n"
            f"{remotes[name]} is what you have."
        )

    cli.success("Done.")



import json
import shutil
import pathlib
import subprocess

from gsc import verifier, cli, setup_exercise
from gsc.exercises import utils

MASTER_COMMIT_MSG = "Fix divide function for some numbers"
BRANCH_COMMIT_MSG = "Fix divide function for all numbers"
FILE_NAME = "useful_things.py"
BRANCH_NAME = "fix-subtract"


def setup():
    state = {}
    # Make sure we're on the master branch
    subprocess.run(["git", "checkout", "master"], capture_output=True)
    # Create branch
    res = subprocess.run(["git", "branch", BRANCH_NAME], capture_output=True)
    if res.returncode != 0:
        raise setup_exercise.SetupError(
            f"Failed to create branch `{BRANCH_NAME}`. Run `gsc reset`."
        )

    codefile = pathlib.Path(FILE_NAME)

    # Create commit on master
    cli.info("Implementing the divide function incorrectly on master.")
    code = codefile.read_text()
    implemented_divide = code.replace(
        """
def divide(x, y):
    return "Hello World"
""",
        """
def divide(x, y):
    # This is probably correct - don't have time to test
    return x % y
""",
    )
    codefile.write_text(implemented_divide)

    res = subprocess.run(["git", "add", FILE_NAME], capture_output=True)
    if res.returncode != 0:
        raise setup_exercise.SetupError("Failed to add code change to master.")

    res = subprocess.run(
        ["git", "commit", "-m", MASTER_COMMIT_MSG], capture_output=True
    )
    if res.returncode != 0:
        raise setup_exercise.SetupError("Failed to commit to master.")

    res = subprocess.run(["git", "rev-parse", "HEAD"], capture_output=True)
    if res.returncode != 0:
        raise setup_exercise.SetupError("Failed to get hash of commit on master.")
    state["master_hash"] = res.stdout.decode("utf-8").strip()

    # Push master branch
    cli.info("Pushing master.")
    res = subprocess.run(["git", "push"], capture_output=True)
    if res.returncode != 0:
        raise setup_exercise.SetupError("Failed to push commit.")

    # Switch branch
    cli.info("Switching branch.")
    subprocess.run(["git", "checkout", BRANCH_NAME], capture_output=True)

    # Create local commit
    cli.info(f"Implementing the divide function correctly on branch `{BRANCH_NAME}`.")
    implemented_divide = code.replace(
        """
def divide(x, y):
    return "Hello World"
""",
        """
def divide(x, y):
    return x / y
""",
    )
    codefile.write_text(implemented_divide)

    res = subprocess.run(["git", "add", FILE_NAME], capture_output=True)
    if res.returncode != 0:
        raise setup_exercise.SetupError("Failed to add code change to branch.")

    res = subprocess.run(
        ["git", "commit", "-m", BRANCH_COMMIT_MSG], capture_output=True
    )
    if res.returncode != 0:
        raise setup_exercise.SetupError("Failed to commit to branch.")

    # Push local branch
    cli.info(f"Pushing {BRANCH_NAME}.")
    res = subprocess.run(
        ["git", "push", "--set-upstream", "origin", BRANCH_NAME], capture_output=True
    )
    if res.returncode != 0:
        raise setup_exercise.SetupError("Failed to push commit.")

    # Back to master
    cli.info("Switching back to master.")
    subprocess.run(["git", "checkout", "master"], capture_output=True)

    # Stash master commit hash for verification later
    pathlib.Path(".gsc_state").write_text(json.dumps(state))

    cli.info(
        "\nUse `git status`, `git log`, and `git branch` to take a look at what's changed in your local repo.\n"
        "When you're ready to start, rebase the `fix-subtract` branch onto master and push it.\n"
    )


def reset():
    subprocess.run(["git", "checkout", "master"], capture_output=True)
    # Reset back to inital commit.
    cli.info("Rewinding history.")
    res = subprocess.run(
        ["git", "rev-list", "--max-parents=0", "HEAD"], capture_output=True
    )
    root_commit = res.stdout.decode("utf-8").strip()
    subprocess.run(["git", "reset", "--hard", root_commit], capture_output=True)
    # Force push master
    cli.info("Resetting remote master.")
    subprocess.run(["git", "push", "--force"], capture_output=True)
    # Delete branch
    cli.info(f"Removing branch `{BRANCH_NAME}`.")
    subprocess.run(["git", "branch", "-D", BRANCH_NAME], capture_output=True)
    # Delete remote branch
    cli.info(f"Removing remote branch `{BRANCH_NAME}`.")
    subprocess.run(["git", "push", "-d", "origin", BRANCH_NAME], capture_output=True)
    # Setup again
    cli.info("Setting up again.")
    setup()


def verify():
    if not utils.clean_status():
        raise verifier.VerifyError(
            "Your git status is not clean. Run `git status` to see the problem."
        )

    state = json.loads(pathlib.Path(".gsc_state").read_text())

    # We should have the branch
    res = subprocess.run(["git", "checkout", BRANCH_NAME], capture_output=True)
    if res.returncode != 0:
        raise verifier.VerifyError(
            f'The "{BRANCH_NAME}" branch has been deleted!\n'
            "Run `gsc reset` to start again."
        )

    # Get commit hashes
    res = subprocess.run(["git", "rev-list", "--all"], capture_output=True)
    commit_hashes = res.stdout.decode("utf-8").strip().split("\n")

    # We should have the master commit
    if state["master_hash"] not in commit_hashes:
        raise verifier.VerifyError(
            f'The "{MASTER_COMMIT_MSG}" commit is missing!\n'
            f"It's on the master branch. You need to bring in into the `{BRANCH_NAME}` branch somehow..."
        )

    commit_messages = utils.commit_messages()
    # We should have the branch commit
    if BRANCH_COMMIT_MSG not in commit_messages:
        raise verifier.VerifyError(
            f'The "{BRANCH_COMMIT_MSG}" commit is missing!\n'
            "Run `gsc reset` and try again."
        )

    # We should not have a merge commit
    if any(["Merge branch" in msg for msg in commit_messages]):
        raise verifier.VerifyError(
            "You created a merge commit when you pulled in the remote commit.\n"
            'Take another look at "Exercise Setup".\n'
            "Run `gsc reset` and try again."
        )

    # We should have chosen the correct version of the code in the conflict
    codefile = pathlib.Path(FILE_NAME)
    code = codefile.read_text()
    if (
        """
def divide(x, y):
    return x / y"""
        not in code
    ):
        raise verifier.VerifyError(
            "You chose the wrong version of the code when fixing the conflict.\n"
            "This implementation is broken!\n"
            'Take another look at "Fix the Merge Conflict".\n'
            "Run `gsc reset` and try again."
        )

    # We should be up to date with the remote.
    utils.assert_up_to_date_with_remote(BRANCH_NAME)

    cli.success("Done.")



import subprocess

from gsc import verifier, cli


def verify():
    res = subprocess.run(["git", "remote", "-v"], capture_output=True)
    res = res.stdout.decode("utf-8").strip()

    if "git@github.com:" not in res:
        raise verifier.VerifyError("Repo not cloned using SSH.")

    cli.success("Looks good.")



import json
import shutil
import pathlib
import subprocess

from gsc import verifier, cli, setup_exercise
from gsc.exercises import utils

REMOTE_COMMIT_MSG = "Implement subtract function"
LOCAL_COMMIT_MSG = "Implement divide function"
FILE_NAME = "useful_things.py"


def setup():
    state = {}
    # Backup .git
    shutil.copytree(".git", ".git.bak")

    codefile = pathlib.Path(FILE_NAME)

    # Create remote commit & push
    cli.info("Implementing the subtract function.")
    code = codefile.read_text()
    implemented_subtract = code.replace(
        """
def subtract(x, y):
    pass
""",
        """
def subtract(x, y):
    return x - y
""",
    )
    codefile.write_text(implemented_subtract)

    res = subprocess.run(["git", "add", FILE_NAME], capture_output=True)
    if res.returncode != 0:
        raise setup_exercise.SetupError("Failed to add first code change.")

    res = subprocess.run(
        ["git", "commit", "-m", REMOTE_COMMIT_MSG], capture_output=True
    )
    if res.returncode != 0:
        raise setup_exercise.SetupError("Failed to commit first code change.")

    res = subprocess.run(["git", "rev-parse", "HEAD"], capture_output=True)
    if res.returncode != 0:
        raise setup_exercise.SetupError("Failed to get hash of first commit.")
    state["remote_hash"] = res.stdout.decode("utf-8").strip()

    cli.info("Pushing the commit.")
    res = subprocess.run(["git", "push"], capture_output=True)
    if res.returncode != 0:
        raise setup_exercise.SetupError("Failed to push commit.")

    # Restore .git
    cli.info("Forgetting about remote commit.")
    utils.rmtree_readonly(".git")
    shutil.copytree(".git.bak", ".git")
    utils.rmtree_readonly(".git.bak")

    # Create local commit
    cli.info("Implementing the divide function.")
    implemented_divide = code.replace(
        """
def divide(x, y):
    pass
""",
        """
def divide(x, y):
    return x / y
""",
    )
    codefile.write_text(implemented_divide)

    res = subprocess.run(["git", "add", FILE_NAME], capture_output=True)
    if res.returncode != 0:
        raise setup_exercise.SetupError("Failed to add second code change.")

    res = subprocess.run(["git", "commit", "-m", LOCAL_COMMIT_MSG], capture_output=True)
    if res.returncode != 0:
        raise setup_exercise.SetupError("Failed to commit second code change.")

    # Stash remote commit hash for verification later
    pathlib.Path(".gsc_state").write_text(json.dumps(state))

    cli.info(
        "\nUse `git status` and `git log` to take a look at what's changed in your local repo.\n"
        "When you're ready to start the exercise, try `git push`.\n"
    )


def reset():
    # Reset back to inital commit.
    cli.info("Rewinding history.")
    res = subprocess.run(
        ["git", "rev-list", "--max-parents=0", "HEAD"], capture_output=True
    )
    root_commit = res.stdout.decode("utf-8").strip()
    subprocess.run(["git", "reset", "--hard", root_commit], capture_output=True)
    # Force push
    cli.info("Cleaning remote repo.")
    subprocess.run(["git", "push", "--force"], capture_output=True)
    # Setup again
    cli.info("Setting up again.")
    setup()


def verify():
    if not utils.clean_status():
        raise verifier.VerifyError(
            "Your git status is not clean. Run `git status` to see the problem."
        )

    if not utils.on_branch("master"):
        raise verifier.VerifyError(
            "You should be on the master branch for this exercise.\n"
            "Change branch with `git checkout master` and try again."
        )

    state = json.loads(pathlib.Path(".gsc_state").read_text())

    # Get commit hashes
    res = subprocess.run(["git", "rev-list", "--all"], capture_output=True)
    commit_hashes = res.stdout.decode("utf-8").strip().split("\n")

    # We should have the remote commit
    if state["remote_hash"] not in commit_hashes:
        raise verifier.VerifyError(
            f'The "{REMOTE_COMMIT_MSG}" commit is missing!\n'
            "It's on the remote repo. You need to bring in into your local somehow..."
        )

    commit_messages = utils.commit_messages()
    # We should have the local commit
    if LOCAL_COMMIT_MSG not in commit_messages:
        raise verifier.VerifyError(
            f'The "{LOCAL_COMMIT_MSG}" commit is missing!\n'
            "Run `gsc reset` and try again."
        )

    # We should not have a merge commit
    if any(["Merge branch" in msg for msg in commit_messages]):
        raise verifier.VerifyError(
            "You created a merge commit when you pulled in the remote commit.\n"
            'Take another look at "How do we fix a sync error?"\n'
            "Run `gsc reset` and try again."
        )

    # We should be up to date with the remote
    utils.assert_up_to_date_with_remote("master")

    cli.success("Done.")



import os
import stat
import shutil
import subprocess
import typing as t
from gsc import cli, verifier


def pluralise_commits(number: str) -> str:
    if number == "1":
        return "is 1 commit"
    else:
        return f"are {number} commits"


def clean_status() -> bool:
    res = subprocess.run(["git", "status", "--porcelain"], capture_output=True)
    return res.stdout.decode("utf-8").strip() == ""


def on_branch(branch: str) -> bool:
    res = subprocess.run(["git", "branch"], capture_output=True)
    return b"* " + branch.encode("utf-8") in res.stdout


def tests_pass() -> bool:
    res = subprocess.run(["pytest"], capture_output=True)
    return res.returncode == 0


def assert_up_to_date_with_remote(branch: str) -> None:
    res = subprocess.run(
        ["git", "rev-list", "--left-right", "--count", f"{branch}...origin/{branch}"],
        capture_output=True,
    )
    commits = res.stdout.decode("utf-8").strip().split("\t")
    if ["0", "0"] != commits:
        raise verifier.VerifyError(
            f"There {pluralise_commits(commits[0])} on your local repo which aren't on your remote.\n"
            f"There {pluralise_commits(commits[1])} on your remote repo which aren't on your local.\n"
            "Your local and remote repos should have the same commits.\n"
            "Run `gsc reset` to try again if you aren't sure what's gone wrong."
        )


def remove_hash(msg: str) -> str:
    chunks = msg.split(" ", 1)[1:]
    return " ".join(chunks).strip()


def most_recent_commit_message() -> str:
    res = subprocess.run(["git", "log", "-1", "--oneline"], capture_output=True)
    return remove_hash(res.stdout.decode("utf-8"))


def most_recent_commit_description() -> str:
    res = subprocess.run(["git", "log", "-1", "--pretty=%B"], capture_output=True)
    commit_msg = res.stdout.decode("utf-8").strip()
    [_title, desc] = commit_msg.split("\n\n", 1)
    return desc


def commit_messages() -> t.List[str]:
    res = subprocess.run(["git", "log", "--oneline"], capture_output=True)
    msgs = res.stdout.decode("utf-8").split("\n")[:-1]
    return list(map(remove_hash, msgs))


def check_commit_message(msg: str) -> None:
    cli.info(f"Checking commit message: {msg}")

    if msg.endswith("."):
        cli.warn(
            "You shouldn't put a full stop at the end of your Git commit messages.\n"
            "Messages should be concise and not gramatically correct sentences."
        )

    if len(msg) >= 50:
        cli.warn(
            "Git commit messages should be fewer than 50 characters long.\n"
            f"You've used {len(msg)}.\n"
            "Try to make your commit messages more concise."
        )


def remove_readonly(func, path, execinfo):
    os.chmod(path, stat.S_IWRITE)
    func(path)


def rmtree_readonly(root):
    shutil.rmtree(root, onerror=remove_readonly)



import os
import pathlib
from gsc import cli, client
from gsc.exercises import (
    my_first_commit,
    push_and_pull,
    ssh_clone,
    sync_error,
    merge_conflict,
    multiple_remotes,
    use_the_force,
    revert,
    reset_file,
    amend,
)


class VerifyError(Exception):
    pass


def verify(exercise: str = None):
    while not os.path.exists(".git"):
        os.chdir("..")
        if os.getcwd() == "/":
            raise VerifyError("This is not a git repo.")

    if exercise == "ssh":
        cli.title(f"Verifying ssh")
        ssh_clone.verify()
        client.complete_exercise("ssh_clone")
        return
    elif exercise:
        raise VerifyError("Unknown Git Scientist exercise. Try upgrading gsc.")

    if os.getcwd().endswith(push_and_pull.PULL_SUFFIX):
        gsc_id = "push_and_pull"
    elif not os.path.exists(".gsc_id"):
        raise VerifyError("This repo is not a Git Scientist exercise.")
    else:
        gsc_id = pathlib.Path(".gsc_id").read_text().strip()

    cli.title(f"Verifying {gsc_id}")

    if gsc_id == "my_first_commit":
        my_first_commit.verify()
    elif gsc_id == "push_and_pull":
        push_and_pull.verify()
    elif gsc_id == "sync_error":
        sync_error.verify()
    elif gsc_id == "merge_conflict":
        merge_conflict.verify()
    elif gsc_id == "multiple_remotes":
        multiple_remotes.verify()
    elif gsc_id == "use_the_force":
        use_the_force.verify()
    elif gsc_id == "revert":
        revert.verify()
    elif gsc_id == "reset_file":
        reset_file.verify()
    elif gsc_id == "amend":
        amend.verify()
    else:
        raise VerifyError("Unknown Git Scientist exercise. Try upgrading gsc.")
    client.complete_exercise(gsc_id)



import os
import typer
import shutil
import subprocess

from gsc import verifier, cli, setup_exercise
from gsc.exercises import utils

COMMIT_MSG = "Setup Git Scientist Exercise"
PULL_SUFFIX = "-gsc-pull"


def setup():
    # Copy repo so we can practice pulling.
    repo_name = os.getcwd()
    pull_repo_name = repo_name + PULL_SUFFIX

    if os.path.exists(pull_repo_name):
        delete = typer.confirm(
            f"The directory {pull_repo_name} already exists. Do you want to delete it?"
        )
        if not delete:
            cli.info("Not deleting.")
            raise typer.Abort()
        cli.info(f"Deleting {pull_repo_name}.")
        utils.rmtree_readonl(pull_repo_name)

    shutil.copytree(repo_name, pull_repo_name)

    # Setup original repo so we can practice pushing.
    with open(".gsc_id", "w") as f:
        f.write("push_and_pull")

    res = subprocess.run(["git", "add", ".gsc_id"], capture_output=True)
    if res.returncode != 0:
        raise setup_exercise.SetupError("Failed to add gsc_id. Contact us for help.")

    res = subprocess.run(["git", "commit", "-m", COMMIT_MSG], capture_output=True)
    if res.returncode != 0:
        raise setup_exercise.SetupError(
            "Failed to setup Git Scientist. Contact us for help."
        )


def verify():
    if not utils.clean_status():
        raise verifier.VerifyError(
            "Your git status is not clean. Run `git status` to see the problem."
        )

    if not utils.on_branch("master"):
        raise verifier.VerifyError(
            "You should be on the master branch for this exercise.\n"
            "Change branch with `git checkout master` and try again."
        )

    cwd = os.getcwd()
    if cwd.endswith(PULL_SUFFIX):
        repo_name = cwd[: -len(PULL_SUFFIX)]
        pull_repo_name = cwd
    else:
        repo_name = cwd
        pull_repo_name = cwd + PULL_SUFFIX

    # Check that both repos contain the commit we want.
    os.chdir(repo_name)
    msgs = utils.commit_messages()
    if COMMIT_MSG not in msgs:
        raise verifier.VerifyError(
            "Setup commit not found. Did you run `gsc setup push_and_pull`?"
        )

    os.chdir(pull_repo_name)
    msgs = utils.commit_messages()
    if COMMIT_MSG not in msgs:
        raise verifier.VerifyError(
            "The commit has not been pulled into your local repo.\n"
            f"The repo is located at {pull_repo_name}\n"
            "See the My First Pull section of the lesson."
        )

    cli.success("Done.")



import shutil
import pathlib
import subprocess

from gsc import verifier, cli, setup_exercise
from gsc.exercises import utils

FILE_NAME = "useful_things.py"
MASTER_COMMIT_MSG = "Fix the subtract function"


def setup():
    # Make sure we're on the master branch
    subprocess.run(["git", "checkout", "master"], capture_output=True)

    # Commit a change which needs to be amended
    codefile = pathlib.Path(FILE_NAME)
    code = codefile.read_text()
    changed = code.replace(
        """
def subtract(x, y):
    return x + y
""",
        """
def subtract(x, y):
    # TODO: delete this comment
    return x - y
""",
    )
    codefile.write_text(changed)

    res = subprocess.run(["git", "add", FILE_NAME], capture_output=True)
    if res.returncode != 0:
        raise setup_exercise.SetupError("Failed to add code change to master.")

    res = subprocess.run(
        ["git", "commit", "-m", MASTER_COMMIT_MSG], capture_output=True
    )
    if res.returncode != 0:
        raise setup_exercise.SetupError("Failed to commit to master.")

    cli.info(
        "\nUse `git status` and `git log` to take a look at what's changed in your local repo.\n"
        "When you're ready to start, amend the most recent commit.\n"
    )


def reset():
    subprocess.run(["git", "checkout", "master"], capture_output=True)
    subprocess.run(["git", "reset", "--hard", "origin/master"], capture_output=True)
    cli.info("Setting up again.")
    setup()


def verify():
    if not utils.clean_status():
        raise verifier.VerifyError(
            "Your git status is not clean. Run `git status` to see the problem."
        )

    # We should have 2 commit
    commit_messages = utils.commit_messages()
    num_commits = len(commit_messages)
    if num_commits != 2:
        raise verifier.VerifyError(
            f"There {utils.pluralise_commits(num_commits)} on your local branch.\n"
            "There should be exactly 2: the inital commit and the amended commit.\n"
            "Run `gsc reset` to try again."
        )

    # The comment should be gone
    codefile = pathlib.Path(FILE_NAME)
    code = codefile.read_text()
    if "TODO" in code:
        raise verifier.VerifyError(
            "You don't seem to have amended the commit.\n"
            "Check the exercise description."
        )

    # The code should be correct
    codefile = pathlib.Path(FILE_NAME)
    code = codefile.read_text()
    if (
        """
def subtract(x, y):
    return x - y"""
        not in code
    ):
        raise verifier.VerifyError(
            "The code is no longer correct.\n"
            "Make sure you don't delete or replace working code when you amend the commit!\n"
            "Run `gsc reset` and try again."
        )

    cli.success("Done.")



import shutil
import pathlib
import subprocess

from gsc import verifier, cli, setup_exercise
from gsc.exercises import utils

REVERT_HASH = "67a6501a99ef70b8d554e2b11357740cb1f9b8a0"
FILE_NAME = "useful_things.py"


def setup():
    cli.info("There's no setup for this exercise.")


def reset():
    subprocess.run(["git", "checkout", "master"], capture_output=True)
    # Reset to origin/master
    subprocess.run(["git", "reset", "--hard", "origin/master"], capture_output=True)


def verify():
    if not utils.clean_status():
        raise verifier.VerifyError(
            "Your git status is not clean. Run `git status` to see the problem."
        )

    commit_messages = utils.commit_messages()

    # We should have 3 commits
    num_commits = len(commit_messages)
    if num_commits != 3:
        raise verifier.VerifyError(
            f"There {utils.pluralise_commits(num_commits)} on your local branch.\n"
            "There should be exactly 3: the 2 original commits and 1 revert commit.\n"
            "Run `gsc reset` to try again."
        )

    # The code should be the same as in the initial commit
    codefile = pathlib.Path(FILE_NAME)
    code = codefile.read_text()
    if (
        """
def subtract(x, y):
    return x + y"""
        not in code
    ):
        raise verifier.VerifyError(
            "You chose the wrong version of the code when fixing the conflict.\n"
            "This implementation is broken!\n"
            'Take another look at "Fix the Merge Conflict".\n'
            "Run `gsc reset` and try again."
        )

    # The revert commit title should start with the word "Revert".
    title = utils.most_recent_commit_message()
    if not title.startswith("Revert"):
        raise verifier.VerifyError(
            'The revert commit title should start with the word "Revert".\n'
            "Git does this automatically when you revert a commit - don't change it!"
        )

    # The reverted commit hash should be in the revert commit description.
    desc = utils.most_recent_commit_description()
    if REVERT_HASH not in desc:
        raise verifier.VerifyError(
            "The commit hash is not in the commit message description.\n"
            "Git puts it there automatically when you revert a commit - don't delete it!"
        )

    cli.success("Done.")



import json
import shutil
import pathlib
import subprocess

from gsc import verifier, cli, setup_exercise
from gsc.exercises import utils

FILE_NAME = "useful_things.py"


def setup():
    # Make sure we're on the master branch
    subprocess.run(["git", "checkout", "master"], capture_output=True)

    # Make an uncommitted file change
    codefile = pathlib.Path(FILE_NAME)
    code = codefile.read_text()
    silly_change = code.replace(
        """
def subtract(x, y):
    return x - y
""",
        """
def subtract(x, y):
    return "Oops I should not have done this"
""",
    )
    codefile.write_text(silly_change)

    state = {"setup": True}
    pathlib.Path(".gsc_state").write_text(json.dumps(state))


def reset():
    subprocess.run(["git", "checkout", "master"], capture_output=True)
    # Reset to origin/master
    subprocess.run(["git", "reset", "--hard", "origin/master"], capture_output=True)
    cli.info("Setting up again.")
    setup()


def verify():
    try:
        pathlib.Path(".gsc_state").read_text()
    except FileNotFoundError:
        raise verifier.VerifyError("You haven't set up the exercise.\nRun `gsc setup`.")

    if not utils.clean_status():
        raise verifier.VerifyError(
            "Your git status is not clean. Run `git status` to see the problem."
        )

    # We should have 1 commit
    commit_messages = utils.commit_messages()
    num_commits = len(commit_messages)
    if num_commits != 1:
        raise verifier.VerifyError(
            f"There {utils.pluralise_commits(num_commits)} on your local branch.\n"
            "There should be exactly 1: the inital commit.\n"
            "Run `gsc reset` to try again."
        )

    # The code should be the same as in the initial commit
    codefile = pathlib.Path(FILE_NAME)
    code = codefile.read_text()
    if (
        """
def subtract(x, y):
    return x - y"""
        not in code
    ):
        raise verifier.VerifyError(
            "The code is not the same as it was originally.\n"
            "You've done something strange. Email us for help if you're not sure what.\n"
            "Run `gsc reset` and try again."
        )

    cli.success("Done.")



import os
import pathlib
from gsc import cli
from gsc.exercises import (
    push_and_pull,
    sync_error,
    merge_conflict,
    multiple_remotes,
    use_the_force,
    revert,
    reset_file,
    amend,
)


class SetupError(Exception):
    pass


def setup(gsc_id: str):
    while not os.path.exists(".git"):
        os.chdir("..")
        if os.getcwd() == "/":
            raise SetupError("This is not a git repo.")

    if not gsc_id:
        if not os.path.exists(".gsc_id"):
            raise SetupError(
                "Could not find a Git Scientist exercise ID.\n"
                "Did you mean to give an ID as an argument to gsc?"
            )
        gsc_id = pathlib.Path(".gsc_id").read_text().strip()

    # Just in case someone decides to use dashes.
    gsc_id = gsc_id.replace("-", "_")

    cli.info(f"Setting up {gsc_id}")

    if gsc_id == "push_and_pull":
        push_and_pull.setup()
    elif gsc_id == "sync_error":
        sync_error.setup()
    elif gsc_id == "merge_conflict":
        merge_conflict.setup()
    elif gsc_id == "multiple_remotes":
        multiple_remotes.setup()
    elif gsc_id == "use_the_force":
        use_the_force.setup()
    elif gsc_id == "revert":
        revert.setup()
    elif gsc_id == "reset_file":
        reset_file.setup()
    elif gsc_id == "amend":
        amend.setup()
    else:
        raise SetupError("Unknown Git Scientist exercise. Try upgrading gsc.")
